@@ -11,6 +11,7 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                                           'nets',
                                           'categorical_columns',
                                           'exclude_columns',
+                                          'task',
                                           'pos_label',
                                           'metrics',
                                           'auto_categorize',
@@ -19,6 +20,7 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                                           'auto_encode_label',
                                           'auto_imputation',
                                           'auto_discrete',
+                                          'auto_discard_unique',
                                           'apply_gbm_features',
                                           'gbm_params',
                                           'gbm_feature_type',
@@ -45,8 +47,10 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                                           'home_dir',
                                           'monitor_metric',
                                           'earlystopping_patience',
+                                          'earlystopping_mode',
                                           'gpu_usage_strategy',
                                           'distribute_strategy',
+                                          'var_len_categorical_columns',
                                           ])):
     def __hash__(self):
         return self.name.__hash__()
@@ -56,6 +60,7 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                 nets=['dnn_nets'],
                 categorical_columns='auto',
                 exclude_columns=[],
+                task=consts.TASK_AUTO,
                 pos_label=None,
                 metrics=['accuracy'],
                 auto_categorize=False,
@@ -64,6 +69,7 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                 auto_encode_label=True,
                 auto_imputation=True,
                 auto_discrete=False,
+                auto_discard_unique=True,
                 apply_gbm_features=False,
                 gbm_params={},
                 gbm_feature_type=consts.GBM_FEATURE_TYPE_EMB,  # embedding/dense
@@ -80,8 +86,8 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                 optimizer='auto',
                 loss='auto',
                 dnn_params={
-                    'dnn_units': ((128, 0, False), (64, 0, False)),
-                    'dnn_activation': 'relu',
+                    'hidden_units': ((128, 0, False), (64, 0, False)),
+                    'activation': 'relu',
                 },
                 autoint_params={
                     'num_attention': 3,
@@ -120,9 +126,26 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                 home_dir=None,
                 monitor_metric=None,
                 earlystopping_patience=1,
+                earlystopping_mode='auto',  # auto,min,max
                 gpu_usage_strategy=consts.GPU_USAGE_STRATEGY_GROWTH,
                 distribute_strategy=None,
+                var_len_categorical_columns=None,
+                # a tuple3, format is (column_name, separator, pool_strategy), pool_strategy is one of max,avg;  e.g. [('genres', '|', 'avg' )]
                 ):
+
+        if var_len_categorical_columns is not None and len(var_len_categorical_columns) > 0:
+            # check items
+            for v in var_len_categorical_columns:
+                _name = v[0]
+                if not isinstance(v, (tuple, list)) or len(v) != 3:
+                    raise ValueError("Var len column config should be a tuple 3.")
+                if exclude_columns is not None:
+                    if _name in exclude_columns:
+                        raise ValueError(f"Var len column {_name} can not put in 'exclude_columns' ")
+                if categorical_columns is not None and isinstance(categorical_columns, list):
+                    if _name in categorical_columns:
+                        raise ValueError(f"Var len column {_name} can not put in 'categorical_columns' ")
+
         nets = deepnets.get_nets(nets)
 
         if home_dir is None and os.environ.get(consts.ENV_DEEPTABLES_HOME) is not None:
@@ -133,6 +156,7 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                                                nets,
                                                categorical_columns,
                                                exclude_columns,
+                                               task,
                                                pos_label,
                                                metrics,
                                                auto_categorize,
@@ -141,6 +165,7 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                                                auto_encode_label,
                                                auto_imputation,
                                                auto_discrete,
+                                               auto_discard_unique,
                                                apply_gbm_features,
                                                gbm_params,
                                                gbm_feature_type,
@@ -167,16 +192,22 @@ class ModelConfig(collections.namedtuple('ModelConfig',
                                                home_dir,
                                                monitor_metric,
                                                earlystopping_patience,
+                                               earlystopping_mode,
                                                gpu_usage_strategy,
                                                distribute_strategy,
+                                               var_len_categorical_columns,
                                                )
+
     @property
     def first_metric_name(self):
-        if self.metrics is None or len(self.metrics)<=0:
+        import tensorflow as tf
+        if self.metrics is None or len(self.metrics) <= 0:
             raise ValueError('`metrics` is none or empty.')
         first_metric = self.metrics[0]
         if isinstance(first_metric, str):
             return first_metric
+        if isinstance(first_metric, tf.metrics.Metric):
+            return first_metric.name
         if callable(first_metric):
             return first_metric.__name__
-        raise ValueError('`metric` must be string or callable object.')
\ No newline at end of file
+        raise ValueError('`metric` must be string or callable object.')
