@@ -3,10 +3,10 @@ import six
 from inspect import signature
 import tensorflow as tf
 from tensorflow.keras.layers import Dense, Concatenate, Flatten, BatchNormalization, Activation, Dropout
-from tensorflow.python.keras.utils.generic_utils import deserialize_keras_object
-from tensorflow.python.keras.utils.generic_utils import serialize_keras_object
+from tensorflow.python.keras.utils.generic_utils import deserialize_keras_object, serialize_keras_object
 
 from . import layers
+from ..utils import counter
 
 WideDeep = ['linear', 'dnn_nets']
 DeepFM = ['linear', 'fm_nets', 'dnn_nets']
@@ -24,6 +24,18 @@ AFM = ['afm_nets']
 #         'autoint_nets', 'fg_nets', 'fgcnn_cin_nets', 'fgcnn_fm_nets', 'fgcnn_ipnn_nets',
 #         'fgcnn_dnn_nets', 'fibi_nets', 'fibi_dnn_nets']
 
+def _concat_embeddings(embeddings, concat_layer_name):
+    if embeddings is not None:
+        len_embeddings = len(embeddings)
+        if len_embeddings > 1:
+            return Concatenate(axis=1, name=concat_layer_name)(embeddings)
+        elif len_embeddings == 1:
+            return embeddings[0]
+        else:
+            return None
+    else:
+        return None
+
 
 def linear(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config, model_desc):
     """
@@ -31,8 +43,8 @@ def linear(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config,
     """
     x = None
     x_emb = None
-    if embeddings is not None and len(embeddings) > 0:
-        concat_embeddings = Concatenate(axis=1, name='concat_linear_embedding')(embeddings)
+    concat_embeddings = _concat_embeddings(embeddings, 'concat_linear_embedding')
+    if concat_embeddings is not None :
         x_emb = tf.reduce_sum(concat_embeddings, axis=-1, name='linear_reduce_sum')
 
     if x_emb is not None and dense_layer is not None:
@@ -56,10 +68,10 @@ def cin_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, confi
     is measured explicitly; (3) the complexity of network will not grow exponentially with the degree
     of interactions.
     """
-    if embeddings is None or len(embeddings) <= 0:
+    cin_concat = _concat_embeddings(embeddings, 'concat_cin_embedding')
+    if cin_concat is None:
         model_desc.add_net('cin', (None), (None))
         return None
-    cin_concat = Concatenate(axis=1, name='concat_cin_embedding')(embeddings)
     cin_output = layers.CIN(params=config.cin_params)(cin_concat)
     model_desc.add_net('cin', cin_concat.shape, cin_output.shape)
     return cin_output
@@ -69,12 +81,14 @@ def fm_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config
     """
     FM models pairwise(order-2) feature interactions
     """
-    if embeddings is None or len(embeddings) <= 0:
+    concat_embeddings_layer = _concat_embeddings(embeddings, 'concat_fm_embedding')
+
+    if concat_embeddings_layer is None :
         model_desc.add_net('fm', (None), (None))
         return None
-    fm_concat = Concatenate(axis=1, name='concat_fm_embedding')(embeddings)
-    fm_output = layers.FM(name='fm_layer')(fm_concat)
-    model_desc.add_net('fm', fm_concat.shape, fm_output.shape)
+    # fm_concat = Concatenate(axis=1, name='concat_fm_embedding')(embeddings)
+    fm_output = layers.FM(name='fm_layer')(concat_embeddings_layer)
+    model_desc.add_net('fm', concat_embeddings_layer.shape, fm_output.shape)
     return fm_output
 
 
@@ -82,7 +96,7 @@ def afm_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, confi
     """Attentional Factorization Machine (AFM), which learns the importance of each feature interaction
     from datasets via a neural attention network.
     """
-    if embeddings is None or len(embeddings) <= 0:
+    if embeddings is None or len(embeddings) < 2:  # Need at last 2 embedding
         return None
     afm_output = layers.AFM(params=config.afm_params, name='afm_layer')(embeddings)
     model_desc.add_net('afm', f'list({len(embeddings)})', afm_output.shape)
@@ -94,6 +108,9 @@ def opnn_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, conf
     Outer Product-based Neural Network
     OuterProduct+DNN
     """
+    if embeddings is None or len(embeddings) < 2:  # at least 2 embedding to build `layers.InnerProduct`
+        return None
+
     op = layers.OuterProduct(config.pnn_params, name='outer_product_layer')(embeddings)
     model_desc.add_net('opnn-outer_product', f'list({len(embeddings)})', op.shape)
 
@@ -108,6 +125,8 @@ def ipnn_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, conf
     Inner Product-based Neural Network
     InnerProduct+DNN
     """
+    if embeddings is None or len(embeddings) < 2:  # at least 2 embedding to build `layers.InnerProduct`
+        return None
 
     ip = layers.InnerProduct(name='inner_product_layer')(embeddings)
     model_desc.add_net('ipnn-inner_product', f'list({len(embeddings)})', ip.shape)
@@ -122,6 +141,8 @@ def pnn_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, confi
     """
     Concatenation of inner product and outer product + DNN
     """
+    if embeddings is None or len(embeddings) < 2: # at least 2 embedding to build `layers.InnerProduct`
+        return None
 
     ip = layers.InnerProduct(name='pnn_inner_product_layer')(embeddings)
     model_desc.add_net('pnn-inner_product', f'list({len(embeddings)})', ip.shape)
@@ -186,15 +207,16 @@ def autoint_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, c
     """
     AutoInt: Automatic Feature Interaction Learning via Self-Attentive Neural Networks.
     """
-    if embeddings is None or len(embeddings) <= 0:
+    concat_embeddings_layer = _concat_embeddings(embeddings, 'concat_autoint_embedding')
+    if concat_embeddings_layer is None:
         model_desc.add_net('autoint', (None), (None))
         return None
-    autoint_emb_concat = Concatenate(axis=1, name='concat_autoint_embedding')(embeddings)
-    output = autoint_emb_concat
+    # autoint_emb_concat = Concatenate(axis=1, name='concat_autoint_embedding')(embeddings)
+    output = concat_embeddings_layer
     for i in range(config.autoint_params['num_attention']):
         output = layers.MultiheadAttention(params=config.autoint_params)(output)
     output = Flatten()(output)
-    model_desc.add_net('autoint', autoint_emb_concat.shape, output.shape)
+    model_desc.add_net('autoint', concat_embeddings_layer.shape, output.shape)
     return output
 
 
@@ -208,10 +230,13 @@ def fg_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config
         .. [1] `Liu B, Tang R, Chen Y, et al. Feature generation by convolutional neural network
         for click-through rate prediction[C]//The World Wide Web Conference. 2019: 1119-1129.`
     """
-    if embeddings is None or len(embeddings) <= 0:
+    fgcnn_emb_concat_index = counter.next_num('concat_fgcnn_embedding')
+    fgcnn_emb_concat = _concat_embeddings(embeddings, f'concat_fgcnn_embedding_{fgcnn_emb_concat_index}')
+
+    if fgcnn_emb_concat is None:
         model_desc.add_net('fgcnn', (None), (None))
         return None
-    fgcnn_emb_concat = Concatenate(axis=1, name='concat_fgcnn_embedding')(embeddings)
+    # fgcnn_emb_concat = Concatenate(axis=1, name=f'concat_fgcnn_embedding_{fgcnn_emb_concat_index}')(embeddings)
     fg_inputs = tf.expand_dims(fgcnn_emb_concat, axis=-1)
     fg_filters = config.fgcnn_params.get('fg_filters', (14, 16))
     fg_heights = config.fgcnn_params.get('fg_heights', (7, 7))
@@ -236,6 +261,10 @@ def fgcnn_cin_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense,
     FGCNN with CIN as deep classifier
     """
     fg_output = fg_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config, model_desc)
+
+    if fg_output is None:
+        return None
+
     cin_output = layers.CIN(params=config.cin_params)(fg_output)
     model_desc.add_net('fgcnn-cin', fg_output.shape, cin_output.shape)
     return cin_output
@@ -245,7 +274,11 @@ def fgcnn_fm_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense,
     """
     FGCNN with FM as deep classifier
     """
+
     fg_output = fg_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config, model_desc)
+    if fg_output is None:
+        return None
+
     fm_output = layers.FM(name='fm_fgcnn_layer')(fg_output)
     model_desc.add_net('fgcnn-fm', fg_output.shape, fm_output.shape)
     return fm_output
@@ -256,6 +289,9 @@ def fgcnn_afm_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense,
     FGCNN with AFM as deep classifier
     """
     fg_output = fg_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config, model_desc)
+    if fg_output is None:
+        return None
+
     split_features = tf.split(fg_output, fg_output.shape[1], axis=1)
     afm_output = layers.AFM(params=config.afm_params)(split_features)
     model_desc.add_net('fgcnn-afm', fg_output.shape, afm_output.shape)
@@ -267,10 +303,15 @@ def fgcnn_ipnn_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense
     FGCNN with IPNN as deep classifier
     """
     fg_output = fg_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config, model_desc)
+    if fg_output is None:
+        return None
     split_features = tf.split(fg_output, fg_output.shape[1], axis=1)
     inner_product = layers.InnerProduct()(split_features)
     # dnn_input = Flatten()(concat_all_features)
-    dnn_input = Concatenate()([Flatten()(fg_output), inner_product, dense_layer])
+    dnn_input_layers = [Flatten()(fg_output), inner_product]
+    if dense_layer is not None:
+        dnn_input_layers.append(dense_layer)
+    dnn_input = Concatenate()(dnn_input_layers)
     dnn_out = dnn(dnn_input, config.dnn_params, cellname='fgcnn_ipnn')
     model_desc.add_net('fgcnn-ipnn', fg_output.shape, dnn_out.shape)
     return dnn_out
@@ -281,7 +322,14 @@ def fgcnn_dnn_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense,
     FGCNN with DNN as deep classifier
     """
     fg_output = fg_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config, model_desc)
-    dnn_input = Concatenate()([Flatten()(fg_output), dense_layer])
+    if fg_output is None:
+        return None
+
+    if dense_layer is not None:
+        dnn_input = Concatenate()([Flatten()(fg_output), dense_layer])
+    else:
+        dnn_input = Flatten()(fg_output)
+
     dnn_out = dnn(dnn_input, config.dnn_params, cellname='fgcnn_dnn')
     model_desc.add_net('fgcnn-ipnn', fg_output.shape, dnn_out.shape)
     return dnn_out
@@ -295,10 +343,11 @@ def fibi_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, conf
     these cross features are concatenated by a combination layer which merges the outputs of
     Bilinear-Interaction layer.
     """
-    if embeddings is None or len(embeddings) <= 0:
+    senet_index = counter.next_num('senet_layer')
+    senet_emb_concat = _concat_embeddings(embeddings, f'concat_senet_embedding_{senet_index}')
+    if senet_emb_concat is None:
         model_desc.add_net('fibi', (None), (None))
         return None
-    senet_emb_concat = Concatenate(axis=1, name='concat_senet_embedding')(embeddings)
 
     senet_pooling_op = config.fibinet_params.get('senet_pooling_op', 'mean')
     senet_reduction_ratio = config.fibinet_params.get('senet_reduction_ratio', 3)
@@ -306,12 +355,12 @@ def fibi_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, conf
 
     senet_embedding = layers.SENET(pooling_op=senet_pooling_op,
                                    reduction_ratio=senet_reduction_ratio,
-                                   name='senet_layer')(senet_emb_concat)
+                                   name=f'senet_layer_{senet_index}')(senet_emb_concat)
     senet_bilinear_out = layers.BilinearInteraction(bilinear_type=bilinear_type,
-                                                    name='senet_bilinear_layer')(senet_embedding)
+                                                    name=f'senet_bilinear_layer_{senet_index}')(senet_embedding)
     bilinear_out = layers.BilinearInteraction(bilinear_type=bilinear_type,
-                                              name='embedding_bilinear_layer')(senet_emb_concat)
-    concat_bilinear = Concatenate(axis=1, name='concat_bilinear')([senet_bilinear_out, bilinear_out])
+                                              name=f'embedding_bilinear_layer_{senet_index}')(senet_emb_concat)
+    concat_bilinear = Concatenate(axis=1, name=f'concat_bilinear_{senet_index}')([senet_bilinear_out, bilinear_out])
     model_desc.add_net('fibi', senet_emb_concat.shape, concat_bilinear.shape)
     return concat_bilinear
 
@@ -320,6 +369,9 @@ def fibi_dnn_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense,
     """
     FiBiNet with DNN as deep classifier
     """
+    if embeddings is None or len(embeddings) <= 1:
+        return None
+
     fibi_output = fibi_nets(embeddings, flatten_emb_layer, dense_layer, concat_emb_dense, config, model_desc)
     dnn_input = Concatenate(name='concat_bilinear_dense')(
         [Flatten(name='flatten_fibi_output')(fibi_output), dense_layer])
