@@ -1,27 +1,31 @@
 # -*- coding:utf-8 -*-
 """Training and inference for tabular datasets using neural nets."""
 
-import datetime
 import os
-
-import numpy as np
-import time
-import pandas as pd
 import pickle
+import time
+
+import dask
+import numpy as np
+import pandas as pd
 from joblib import Parallel, delayed
 from sklearn.metrics import roc_auc_score
-from sklearn.model_selection import KFold, StratifiedKFold, train_test_split
+from sklearn.model_selection import KFold, StratifiedKFold
 from tensorflow.keras.callbacks import EarlyStopping
 from tensorflow.keras.layers import Concatenate, BatchNormalization
-from tensorflow.keras.utils import to_categorical
 
+from hypernets.tabular import get_tool_box, is_dask_installed
 from . import modelset, deepnets
 from .config import ModelConfig
 from .deepmodel import DeepModel
-from .preprocessor import DefaultPreprocessor
-from ..utils import dt_logging, consts
+from .preprocessor import DefaultPreprocessor, DefaultDaskPreprocessor
+from ..utils import dt_logging, consts, fs
+from ..utils.tf_version import tf_less_than
 
-logger = dt_logging.get_logger()
+if is_dask_installed:
+    from hypernets.tabular.dask_ex import DaskToolBox
+
+logger = dt_logging.get_logger(__name__)
 
 
 class DeepTable:
@@ -101,9 +105,9 @@ class DeepTable:
 
             metrics: list of string or callable object, (default=['accuracy'])
                 List of metrics to be evaluated by the model during training and testing.
-                Typically you will use `metrics=['accuracy']` or `metrics=['AUC']`.
-                Every metric should be a built-in evaluation metric in tf.keras.metrics or a callable object
-                like `r2(y_true, y_pred):...` .
+                Typically, you will use `metrics=['accuracy']` or `metrics=['AUC']`.
+                Every metric should be a built-in evaluation metric in tf.keras.metrics or a tf.keras.metrics Object
+                or a callable object like `r2(y_true, y_pred):...` .
                 See also: https://tensorflow.google.cn/versions/r2.0/api_docs/python/tf/keras/metrics
 
             auto_categorize: bool, (default=False)
@@ -157,7 +161,7 @@ class DeepTable:
 
             loss: str or object, (default='auto')
 
-            dnn_params: dict, (default={'dnn_units': ((128, 0, False), (64, 0, False)),
+            dnn_params: dict, (default={'hidden_units': ((128, 0, False), (64, 0, False)),
                                         'dnn_activation': 'relu'})
 
             autoint_params:dict, (default={'num_attention': 3,'num_heads': 1,
@@ -275,15 +279,17 @@ class DeepTable:
     >>>preds = dt.predict(X_eval)
     """
 
-    def __init__(self, config=None):
+    def __init__(self, config=None, preprocessor=None):
         if config is None:
             config = ModelConfig()
         self.config = config
         self.nets = config.nets
         self.output_path = self._prepare_output_dir(config.home_dir, self.nets)
-        self.preprocessor = DefaultPreprocessor(config)
+
+        self.preprocessor = preprocessor
         self.__current_model = None
-        self.__modelset = modelset.ModelSet(metric=self.config.first_metric_name, best_mode=consts.MODEL_SELECT_MODE_AUTO)
+        self.__modelset = modelset.ModelSet(metric=self.config.first_metric_name,
+                                            best_mode=consts.MODEL_SELECT_MODE_AUTO)
 
     @property
     def task(self):
@@ -293,6 +299,10 @@ class DeepTable:
     def num_classes(self):
         return len(self.preprocessor.labels)
 
+    @property
+    def classes_(self):
+        return self.preprocessor.labels
+
     @property
     def pos_label(self):
         if self.config.pos_label is not None:
@@ -327,24 +337,29 @@ class DeepTable:
             max_queue_size=10, workers=1, use_multiprocessing=False):
         logger.info(f'X.Shape={np.shape(X)}, y.Shape={np.shape(y)}, batch_size={batch_size}, config={self.config}')
         logger.info(f'metrics:{self.config.metrics}')
+        if np.ndim(X) != 2:
+            raise ValueError("Input train data should be 2d .")
+        X_shape = np.shape(X)
+        if X_shape[1] < 1:
+            raise ValueError("Input train data should has 1 feature at least.")
         self.__modelset.clear()
 
+        if self.preprocessor is None:
+            self.preprocessor = _get_default_preprocessor(self.config, X, y)
+
         X, y = self.preprocessor.fit_transform(X, y)
         if validation_data is not None:
-            if len(validation_data) != 2:
-                raise ValueError(f'Unexpected validation_data length, expected 2 but {len(validation_data)}.')
-            X_val, y_val = validation_data(0), validation_data(1)
-            X_val, y_val = self.preprocessor.transform(X_val, y_val)
-            validation_data = (X_val, y_val)
+            validation_data = self.preprocessor.transform(*validation_data)
 
-        logger.info(f'training...')
+        logger.info(f'Training...')
         if class_weight is None and self.config.apply_class_weight and self.task != consts.TASK_REGRESSION:
             class_weight = self.get_class_weight(y)
 
         callbacks = self.__inject_callbacks(callbacks)
         model = DeepModel(self.task, self.num_classes, self.config,
                           self.preprocessor.categorical_columns,
-                          self.preprocessor.continuous_columns)
+                          self.preprocessor.continuous_columns,
+                          var_categorical_len_columns=self.preprocessor.var_len_categorical_columns)
         history = model.fit(X, y, batch_size=batch_size, epochs=epochs, verbose=verbose, shuffle=shuffle,
                             validation_split=validation_split, validation_data=validation_data,
                             validation_steps=validation_steps, validation_freq=validation_freq,
@@ -360,38 +375,56 @@ class DeepTable:
                              batch_size=None, epochs=1, verbose=1, callbacks=None, n_jobs=1, random_state=9527,
                              shuffle=True, class_weight=None, sample_weight=None,
                              initial_epoch=0, steps_per_epoch=None, validation_steps=None, validation_freq=1,
-                             max_queue_size=10, workers=1, use_multiprocessing=False
+                             max_queue_size=10, workers=1, use_multiprocessing=False,
+                             oof_metrics=None,
                              ):
-        print("Start cross validation")
         start = time.time()
+        logger.info("Start cross validation")
         logger.info(f'X.Shape={np.shape(X)}, y.Shape={np.shape(y)}, batch_size={batch_size}, config={self.config}')
         logger.info(f'metrics:{self.config.metrics}')
         self.__modelset.clear()
 
+        if self.preprocessor is None:
+            self.preprocessor = _get_default_preprocessor(self.config, X, y)
+
         X, y = self.preprocessor.fit_transform(X, y)
 
         if X_eval is not None:
-            print(f'transform X_eval')
+            logger.info(f'transform X_eval')
             X_eval = self.preprocessor.transform_X(X_eval)
         if X_test is not None:
-            print(f'transform X_test')
+            logger.info(f'transform X_test')
             X_test = self.preprocessor.transform_X(X_test)
 
         if iterators is None:
-            if stratified and self.task != consts.TASK_REGRESSION:
+            if is_dask_installed and DaskToolBox.exist_dask_object(X, y):
+                iterators = KFold(n_splits=num_folds, shuffle=True, random_state=random_state)
+            elif stratified and self.task != consts.TASK_REGRESSION:
                 iterators = StratifiedKFold(n_splits=num_folds, shuffle=True, random_state=random_state)
             else:
                 iterators = KFold(n_splits=num_folds, shuffle=True, random_state=random_state)
-        print(f'Iterators:{iterators}')
+        logger.info(f'Iterators:{iterators}')
 
+        tb = get_tool_box(X)
+        X_shape = tb.get_shape(X)
         test_proba_mean = None
         eval_proba_mean = None
-        if self.task == consts.TASK_MULTICLASS:
-            oof_proba = np.zeros((y.shape[0], self.num_classes))
+        if self.task in (consts.TASK_MULTICLASS, consts.TASK_MULTILABEL):
+            oof_proba = np.full((X_shape[0], self.num_classes), np.nan)
         else:
-            oof_proba = np.zeros((y.shape[0], 1))
+            oof_proba = np.full((X_shape[0], 1), np.nan)
+
+        if is_dask_installed and DaskToolBox.exist_dask_object(X, y):
+            X = DaskToolBox.reset_index(DaskToolBox.to_dask_frame_or_series(X))
+            y = DaskToolBox.to_dask_type(y)
+            if DaskToolBox.is_dask_dataframe_or_series(y):
+                y = y.to_dask_array(lengths=True)
+            X, y = dask.persist(X, y)
+            to_split = np.arange(X_shape[0]), None
+        else:
+            y = np.array(y)
+            to_split = X, y
 
-        y = np.array(y)
         if class_weight is None and self.config.apply_class_weight and self.task == consts.TASK_BINARY:
             class_weight = self.get_class_weight(y)
 
@@ -406,15 +439,17 @@ class DeepTable:
             initial_epoch=initial_epoch, steps_per_epoch=steps_per_epoch,
             max_queue_size=max_queue_size, workers=workers, use_multiprocessing=use_multiprocessing
         )
+        oof_scores = [] if oof_metrics is not None else None
+
         with parallel:
             out = parallel(delayed(_fit_and_score)(
                 self.task, self.num_classes, self.config,
                 self.preprocessor.categorical_columns, self.preprocessor.continuous_columns,
                 n_fold, valid_idx,
-                X.iloc[train_idx], y[train_idx], X.iloc[valid_idx], y[valid_idx],
+                tb.select_df(X, train_idx), y[train_idx], tb.select_df(X, valid_idx), y[valid_idx],
                 X_eval, X_test, f'{self.output_path}{"_".join(self.nets)}-kfold-{n_fold + 1}.h5',
                 **fit_and_score_kwargs)
-                           for n_fold, (train_idx, valid_idx) in enumerate(iterators.split(X, y)))
+                           for n_fold, (train_idx, valid_idx) in enumerate(iterators.split(*to_split)))
 
             for n_fold, idx, history, fold_oof_proba, fold_eval_proba, fold_test_proba in out:
                 oof_proba[idx] = fold_oof_proba
@@ -428,31 +463,81 @@ class DeepTable:
                         test_proba_mean = fold_test_proba / num_folds
                     else:
                         test_proba_mean += fold_test_proba / num_folds
+                if oof_metrics is not None:
+                    fold_y_true = y[idx]
+                    if self.task == consts.TASK_BINARY:
+                        fold_y_proba = self._fix_softmax_proba(fold_oof_proba.shape[0], fold_oof_proba.copy())
+                    else:
+                        fold_y_proba = fold_oof_proba.copy()
+                    fold_y_true = self.preprocessor.inverse_transform_y(fold_y_true)
+                    fold_y_pred = self.proba2predict(fold_y_proba, encode_to_label=True)
+                    oof_scores.append(tb.metrics.calc_score(fold_y_true, fold_y_pred, fold_y_proba, task=self.task,
+                                                            metrics=oof_metrics, pos_label=self.pos_label,
+                                                            classes=self.classes_))
+
                 self.__push_model('val', f'{"+".join(self.nets)}-kfold-{n_fold + 1}',
                                   f'{self.output_path}{"_".join(self.nets)}-kfold-{n_fold + 1}.h5', history)
+        oof_proba_origin = oof_proba.copy()
+        nan_idx = np.argwhere(np.isnan(oof_proba).any(1)).ravel()
+        if self.task == consts.TASK_BINARY:
+            oof_proba_fixed = self._fix_softmax_proba(X.shape[0], oof_proba_origin.copy())
+        elif self.task == consts.TASK_REGRESSION:
+            oof_proba_fixed = oof_proba_origin.reshape(X.shape[0])
+        else:
+            oof_proba_fixed = oof_proba_origin
 
-        if oof_proba.shape[-1] == 1:
-            oof_proba = oof_proba.reshape(-1)
-        if eval_proba_mean is not None and eval_proba_mean.shape[-1] == 1:
-            eval_proba_mean = eval_proba_mean.reshape(-1)
-        if test_proba_mean is not None and test_proba_mean.shape[-1] == 1:
-            test_proba_mean = test_proba_mean.reshape(-1)
-            file = f'{self.output_path}{"_".join(self.nets)}-cv-{num_folds}.csv'
-            pd.DataFrame(test_proba_mean).to_csv(file, index=False)
-        print(f'fit_cross_validation cost:{time.time() - start}')
-        return oof_proba, eval_proba_mean, test_proba_mean
+        if len(nan_idx) > 0:
+            oof_proba_fixed[nan_idx] = np.nan
 
-    def evaluate(self, X_test, y_test, batch_size=256, verbose=0, model_selector=consts.MODEL_SELECTOR_CURRENT, ):
+        if eval_proba_mean is not None:
+            if self.task == consts.TASK_BINARY:
+                eval_proba_mean_fixed = self._fix_softmax_proba(X_eval.shape[0], eval_proba_mean.copy())
+            else:
+                eval_proba_mean_fixed = eval_proba_mean
+        else:
+            eval_proba_mean_fixed = eval_proba_mean
+
+        if test_proba_mean is not None:
+            if self.task == consts.TASK_BINARY:
+                test_proba_mean_fixed = self._fix_softmax_proba(X_test.shape[0], test_proba_mean.copy())
+                file = f'{self.output_path}{"_".join(self.nets)}-cv-{num_folds}.csv'
+                with fs.open(file, 'w', encoding='utf-8') as f:
+                    pd.DataFrame(test_proba_mean.reshape(-1)).to_csv(f, index=False)
+
+            else:
+                test_proba_mean_fixed = test_proba_mean
+        else:
+            test_proba_mean_fixed = test_proba_mean
+
+        logger.info(f'fit_cross_validation taken {time.time() - start}s')
+
+        if oof_metrics is not None:
+            return oof_proba_fixed, eval_proba_mean_fixed, test_proba_mean_fixed, oof_scores
+        else:
+            return oof_proba_fixed, eval_proba_mean_fixed, test_proba_mean_fixed
+
+    def _fix_softmax_proba(self, n_rows, proba):
+        # proba shape should be (n, 1) if output layer is softmax
+        if proba is None:
+            return None
+        else:
+            # assert proba.shape == (n_rows, 1)
+            # return np.insert(proba, 0, values=(1 - proba).reshape(1, -1), axis=1)
+            return get_tool_box(proba).fix_binary_predict_proba_result(proba)
+
+    def evaluate(self, X_test, y_test, batch_size=256, verbose=0,
+                 model_selector=consts.MODEL_SELECTOR_CURRENT, return_dict=True):
         X_t, y_t = self.preprocessor.transform(X_test, y_test)
-        y_t = np.array(y_t)
+        # y_t = np.array(y_t)
         model = self.get_model(model_selector)
         if not isinstance(model, DeepModel):
             raise ValueError(f'Wrong model_selector:{model_selector}')
-        result = model.evaluate(X_t, y_t, batch_size=batch_size, verbose=verbose)
+        result = model.evaluate(X_t, y_t, batch_size=batch_size, verbose=verbose, return_dict=return_dict)
         return result
 
     def predict_proba(self, X, batch_size=128, verbose=0,
                       model_selector=consts.MODEL_SELECTOR_CURRENT, auto_transform_data=True, ):
+        n_rows = X.shape[0]
         start = time.time()
         if model_selector == consts.MODEL_SELECTOR_ALL:
             models = self.get_model(model_selector)
@@ -465,15 +550,14 @@ class DeepTable:
                     proba_avg = np.zeros(proba.shape)
                 proba_avg += proba
             proba_avg /= len(models)
-            print(f'predict_proba cost:{time.time() - start}')
-            return proba_avg
+            proba = proba_avg
         else:
             proba = self.__predict(self.get_model(model_selector),
                                    X, batch_size=batch_size,
                                    verbose=verbose,
                                    auto_transform_data=auto_transform_data)
-            print(f'predict_proba cost:{time.time() - start}')
-            return proba
+        logger.info(f'predict_proba taken {time.time() - start}s')
+        return proba
 
     def predict_proba_all(self, X, batch_size=128, verbose=0, auto_transform_data=True, ):
         mis = self.__modelset.get_modelinfos()
@@ -522,7 +606,7 @@ class DeepTable:
         if auto_transform_data:
             X = self.preprocessor.transform_X(X)
         output = model.apply(X, output_layers, concat_outputs, batch_size, verbose, transformer)
-        print(f'apply cost:{time.time() - start}')
+        logger.info(f'apply taken {time.time() - start}s')
         return output
 
     def concat_emb_dense(self, flatten_emb_layer, dense_layer):
@@ -536,7 +620,7 @@ class DeepTable:
         else:
             raise ValueError('No input layer exists.')
         x = BatchNormalization(name='bn_concat_emb_dense')(x)
-        print(f'Concat embedding and dense layer shape:{x.shape}')
+        logger.info(f'Concat embedding and dense layer shape:{x.shape}')
         return x
 
     def get_model(self, model_selector=consts.MODEL_SELECTOR_CURRENT, ):
@@ -565,18 +649,23 @@ class DeepTable:
         return mi.model
 
     def get_class_weight(self, y):
-        print('Calc classes weight.')
-        if len(y.shape) == 1:
-            y = to_categorical(y)
-        y_sum = y.sum(axis=0)
-        class_weight = {}
-        total = y.shape[0]
-        classes = len(y_sum)
-        print(f"Examples:\nTotal:{total}")
-        for i in range(classes):
-            weight = total / y_sum[i] / classes
-            class_weight[i] = weight
-            print(f'class {i}:{weight}')
+        # logger.info('Calc classes weight.')
+        # if len(y.shape) == 1:
+        #     y = to_categorical(y)
+        # y_sum = y.sum(axis=0)
+        # class_weight = {}
+        # total = y.shape[0]
+        # classes = len(y_sum)
+        # logger.info(f"Examples:\nTotal:{total}")
+        # for i in range(classes):
+        #     weight = total / y_sum[i] / classes
+        #     class_weight[i] = weight
+        #     logger.info(f'class {i}:{weight}')
+
+        n = len(self.classes_)
+        class_weight = get_tool_box(y).compute_class_weight('balanced', classes=range(n), y=y)
+        class_weight = {k: v for k, v in zip(range(n), class_weight)}
+        logger.info(f'classes weight: {class_weight}')
 
         return class_weight
 
@@ -586,17 +675,22 @@ class DeepTable:
         if home_dir[-1] == '/':
             home_dir = home_dir[:-1]
 
-        running_dir = f'dt_{datetime.datetime.now().__format__("%Y%m%d %H%M%S")}_{"_".join(nets)}'
+        running_dir = f'dt_{time.strftime("%Y%m%d%H%M%S")}_{"_".join(nets)}'
         output_path = os.path.expanduser(f'{home_dir}/{running_dir}/')
-        if not os.path.exists(output_path):
-            os.makedirs(output_path)
+        if not fs.exists(output_path):
+            fs.makedirs(output_path, exist_ok=True)
         return output_path
 
     def __predict(self, model, X, batch_size=128, verbose=0, auto_transform_data=True, ):
         logger.info("Perform prediction...")
         if auto_transform_data:
             X = self.preprocessor.transform_X(X)
-        return model.predict(X, batch_size=batch_size, verbose=verbose)
+        proba = model.predict(X, batch_size=batch_size, verbose=verbose)
+        if self.task == consts.TASK_BINARY:
+            # return self._fix_softmax_proba(X.shape[0], proba)
+            return get_tool_box(proba).fix_binary_predict_proba_result(proba)
+        else:
+            return proba
 
     def __set_model(self, type, name, model, history):
         self.__modelset.clear()
@@ -607,7 +701,7 @@ class DeepTable:
         if save_model and isinstance(model, DeepModel):
             modelfile = f'{self.output_path}{name}.h5'
             model.save(modelfile)
-            print(f'Model has been saved to:{modelfile}')
+            logger.info(f'Model has been saved to:{modelfile}')
         mi = modelset.ModelInfo(type, name, model, {}, history=history, modelfile=modelfile)
         self.__modelset.push(mi)
         self.__current_model = mi.name
@@ -624,10 +718,14 @@ class DeepTable:
         else:
             callbacks = []
 
-        if 'auc' in self.monitor.lower() or 'acc' in self.monitor.lower():
-            mode = 'max'
+        if self.config.earlystopping_mode == 'auto':
+            if self.monitor.lower() in ['auc', 'acc', 'accuracy', 'precision', 'recall', 'f1',
+                                        'val_auc', 'val_acc', 'val_accuracy', 'val_precision', 'val_recall', 'val_f1']:
+                mode = 'max'
+            else:
+                mode = 'min'
         else:
-            mode = 'min'
+            mode = self.config.earlystopping_mode
         # if mcp is None:
         #    mcp = ModelCheckpoint(self.model_filepath,
         #                          monitor=self.monitor,
@@ -638,45 +736,69 @@ class DeepTable:
         #                          save_freq='epoch',
         #                          )
         #    callbacks.append(mcp)
-        #    print(f'Injected a callback [ModelCheckpoint].\nfilepath:{mcp.filepath}\nmonitor:{mcp.monitor}')
-        if es is None:
-            es = EarlyStopping(monitor=self.monitor,
+        #    logger.info(f'Injected a callback [ModelCheckpoint].\nfilepath:{mcp.filepath}\nmonitor:{mcp.monitor}')
+        es_patience = self.config.earlystopping_patience
+        if es is None and isinstance(es_patience, int) and es_patience > 0:
+            es = EarlyStopping(monitor=self.monitor if tf_less_than('2.2') else self.monitor.lower(),
                                restore_best_weights=True,
-                               patience=self.config.earlystopping_patience,
+                               patience=es_patience,
                                verbose=1,
-                               min_delta=0.001,
+                               # min_delta=0.0001,
                                mode=mode,
                                baseline=None,
                                )
             callbacks.append(es)
-            print(f'Injected a callback [EarlyStopping]. monitor:{es.monitor}, patience:{es.patience}, mode:{mode}')
+            if logger.is_info_enabled():
+                logger.info(f'Injected a callback [EarlyStopping]. monitor:{es.monitor}, '
+                            f'patience:{es.patience}, mode:{mode}')
         return callbacks
 
-    def save(self, filepath):
+    def __getstate__(self):
+        import copy
+
+        try:
+            state = super().__getstate__()
+        except AttributeError:
+            state = self.__dict__.copy()
+
+        if self.config.distribute_strategy is not None:
+            tmp_conf = self.config._replace(distribute_strategy=None)
+            tmp_preprocessor = copy.deepcopy(self.preprocessor)
+            tmp_preprocessor.config = self.preprocessor.config._replace(distribute_strategy=None)
+            state['config'] = tmp_conf
+            state['preprocessor'] = tmp_preprocessor
+
+        return state
+
+    def save(self, filepath, deepmodel_basename=None):
         if filepath[-1] != '/':
             filepath = filepath + '/'
 
-        if not os.path.exists(filepath):
-            os.mkdirs(filepath)
-
+        if not fs.exists(filepath):
+            fs.makedirs(filepath, exist_ok=True)
+        num_model = len(self.__modelset.get_modelinfos())
         for mi in self.__modelset.get_modelinfos():
             if isinstance(mi.model, str):
                 dm = self.load_deepmodel(mi.model)
                 mi.model = dm
             if not isinstance(mi.model, DeepModel):
                 raise ValueError(f'Currently does not support saving non-DeepModel models.')
+
+            if num_model == 1 and deepmodel_basename is not None:
+                mi.name = deepmodel_basename
+                self.__current_model = deepmodel_basename
             modelfile = f'{filepath}{mi.name}.h5'
             mi.model.save(modelfile)
             mi.model = modelfile
 
-        with open(f'{filepath}dt.pkl', 'wb') as output:
-            pickle.dump(self, output, protocol=2)
+        with fs.open(f'{filepath}dt.pkl', 'wb') as output:
+            pickle.dump(self, output, protocol=4)
 
     @staticmethod
     def load(filepath):
         if filepath[-1] != '/':
             filepath = filepath + '/'
-        with open(f'{filepath}dt.pkl', 'rb') as input:
+        with fs.open(f'{filepath}dt.pkl', 'rb') as input:
             dt = pickle.load(input)
             dt.restore_modelset(filepath)
             return dt
@@ -690,10 +812,11 @@ class DeepTable:
                 mi.model = dm
 
     def load_deepmodel(self, filepath):
-        if os.path.exists(filepath):
-            print(f'Load model from disk:{filepath}.')
+        if fs.exists(filepath):
+            logger.info(f'Load model from: {filepath}.')
             dm = DeepModel(self.task, self.num_classes, self.config,
-                           self.preprocessor.categorical_columns, self.preprocessor.continuous_columns, filepath)
+                           self.preprocessor.categorical_columns,
+                           self.preprocessor.continuous_columns, model_file=filepath)
             return dm
         else:
             raise ValueError(f'Invalid model filename:{filepath}.')
@@ -706,7 +829,7 @@ def _fit_and_score(task, num_classes, config, categorical_columns, continuous_co
                    shuffle=True, class_weight=None, sample_weight=None,
                    initial_epoch=0, steps_per_epoch=None, validation_steps=None, validation_freq=1,
                    max_queue_size=10, workers=1, use_multiprocessing=False):
-    print(f'\nFold:{n_fold + 1}\n')
+    logger.info(f'\nFold:{n_fold + 1}\n')
     model = DeepModel(task, num_classes, config, categorical_columns, continuous_columns)
     history = model.fit(X_train, y_train, batch_size=batch_size, epochs=epochs, verbose=verbose,
                         callbacks=callbacks, validation_data=(X_val, y_val),
@@ -714,59 +837,32 @@ def _fit_and_score(task, num_classes, config, categorical_columns, continuous_co
                         initial_epoch=initial_epoch, steps_per_epoch=steps_per_epoch,
                         validation_steps=validation_steps, validation_freq=validation_freq,
                         max_queue_size=max_queue_size, workers=workers, use_multiprocessing=use_multiprocessing)
-    print(f'Fold {n_fold + 1} fitting over.')
+    logger.info(f'Fold {n_fold + 1} fitting over.')
+
     oof_proba = model.predict(X_val)
-    eval_proba = None
-    test_proba = None
-    if X_eval is not None:
-        eval_proba = model.predict(X_eval)
-    if X_test is not None:
-        test_proba = model.predict(X_test)
-        if model_file is not None:
-            file = f'{model_file}.test_proba.csv'
-            pd.DataFrame(test_proba).to_csv(file, index=False)
-    print(f'Fold {n_fold + 1} scoring over.')
+    eval_proba = model.predict(X_eval) if X_eval is not None else None
+    test_proba = model.predict(X_test) if X_test is not None else None
+    oof_proba, eval_proba, test_proba = get_tool_box(X_val).to_local(oof_proba, eval_proba, test_proba)
+    logger.info(f'Fold {n_fold + 1} scoring over.')
+
     if model_file is not None:
         model.save(model_file)
-        print(f'Save model to:{model_file}.')
+        logger.info(f'Save model to:{model_file}.')
+
+        if X_test is not None:
+            file = f'{model_file}.test_proba.csv'
+            with fs.open(file, 'w', encoding='utf-8') as f:
+                pd.DataFrame(test_proba).to_csv(f, index=False)
+
     model.release()
     return n_fold, valid_idx, history.history, oof_proba, eval_proba, test_proba
 
 
-def infer_task_type(y):
-    uniques = set(y)
-    n_unique = len(uniques)
-    labels = []
-
-    if n_unique == 2:
-        print(f'2 class detected, {uniques}, so inferred as a [binary classification] task')
-        task = consts.TASK_BINARY  # TASK_BINARY
-        labels = sorted(uniques)
-    else:
-        if y.dtype == 'float':
-            print(f'Target column type is float, so inferred as a [regression] task.')
-            task = consts.TASK_REGRESSION
-        else:
-            if n_unique > 1000:
-                if 'int' in y.dtype:
-                    print(
-                        'The number of classes exceeds 1000 and column type is int, so inferred as a [regression] task ')
-                    task = consts.TASK_REGRESSION
-                else:
-                    raise ValueError(
-                        'The number of classes exceeds 1000, please confirm whether your predict target is correct ')
-            else:
-                print(f'{n_unique} class detected, inferred as a [multiclass classification] task')
-                task = consts.TASK_MULTICLASS
-                labels = sorted(uniques)
-    return task, labels
-
-
 def probe_evaluate(dt, X, y, X_test, y_test, layers, score_fn={}):
     from sklearn.linear_model import LogisticRegression
-    print('Extracting features of train set...')
+    logger.info('Extracting features of train set...')
     features_train = dt.apply(X, output_layers=layers)
-    print('Extracting features of test set...')
+    logger.info('Extracting features of test set...')
     features_test = dt.apply(X_test, output_layers=layers)
     y = dt.preprocessor.transform_y(y)
     y_test = dt.preprocessor.transform_y(y_test)
@@ -778,24 +874,31 @@ def probe_evaluate(dt, X, y, X_test, y_test, layers, score_fn={}):
     result = {}
     for i, x_train in enumerate(features_train):
         clf = LogisticRegression(random_state=0).fit(x_train, y)
-        print(f'Fit model for layer[{layers[i]}]...')
+        logger.info(f'Fit model for layer[{layers[i]}]...')
         y_proba = clf.predict_proba(features_test[i])[:, 1]
         y_score = clf.predict(features_test[i])
-        print(f'Scoring...')
+        logger.info(f'Scoring...')
         if len(score_fn) == 0:
             score = clf.score(features_test[i], y_test)
-            print(f'Evaluating accuracy score...')
+            logger.info(f'Evaluating accuracy score...')
             result[layers[i]] = {'accuracy': score}
         else:
             result[layers[i]] = {}
             for metric in score_fn.keys():
-                print(f'Evaluating {metric} score...')
+                logger.info(f'Evaluating {metric} score...')
                 fn = score_fn[metric]
                 if fn == roc_auc_score:
                     score = fn(y_test, y_proba)
                 else:
                     score = fn(y_test, y_score)
                 result[layers[i]][metric] = score
-                print(f'{metric}:{score}')
+                logger.info(f'{metric}:{score}')
             # result[layers[i]] = {metric:score_fn[metric](features_test[i], y_score) for metric in score_fn.keys()}
     return result
+
+
+def _get_default_preprocessor(config, X, y):
+    if is_dask_installed and DaskToolBox.exist_dask_object(X, y):
+        return DefaultDaskPreprocessor(config, compute_to_local=False)
+    else:
+        return DefaultPreprocessor(config)
