@@ -15,7 +15,7 @@ import itertools
 
 from ..utils import dt_logging, consts, gpu
 
-logger = dt_logging.get_logger()
+logger = dt_logging.get_logger(__name__)
 
 class FM(Layer):
     """Factorization Machine to model order-2 feature interactions
@@ -204,7 +204,8 @@ class FGCNN(Layer):
             kernel_initializer='glorot_uniform',
             use_bias=True
         )
-        self.maxpooling2d = MaxPooling2D(pool_size=(self.pool_height, 1))
+        # set padding to prevent dimension during downsampling
+        self.maxpooling2d = MaxPooling2D(pool_size=(self.pool_height, 1), padding='same')
         self.flatten = Flatten()
         self.dense_output = Dense(
             units=input_shape[1] * input_shape[2] * self.new_filters,
@@ -226,9 +227,9 @@ class FGCNN(Layer):
 
     def get_config(self, ):
         config = {'filters': self.filters,
-                  'kernel_width': self.kernel_height,
+                  'kernel_height': self.kernel_height,
                   'new_filters': self.new_filters,
-                  'pool_width': self.pool_height,
+                  'pool_height': self.pool_height,
                   'activation': self.activation, }
         base_config = super(FGCNN, self).get_config()
         return dict(list(base_config.items()) + list(config.items()))
@@ -912,6 +913,57 @@ class MultiColumnEmbedding(Layer):
         return dict(list(base_config.items()) + list(config.items()))
 
 
+class VarLenColumnEmbedding(Embedding):
+    def __init__(self, pooling_strategy='max', dropout_rate=0.,  **kwargs):
+        if pooling_strategy not in ['mean', 'max']:
+            raise ValueError("Param strategy should is one of mean, max")
+        self.pooling_strategy = pooling_strategy
+        self.dropout_rate = dropout_rate  # 支持dropout
+        super(VarLenColumnEmbedding, self).__init__(**kwargs)
+
+    def build(self, input_shape):
+        super(VarLenColumnEmbedding, self).build(input_shape)  # Be sure to call this somewhere!
+        height = input_shape[1]
+        if self.pooling_strategy == "mean":
+            self._pooling_layer = tf.keras.layers.AveragePooling2D(pool_size=(height, 1))
+        else:
+            self._pooling_layer = tf.keras.layers.MaxPooling2D(pool_size=(height, 1))
+
+        if self.dropout_rate > 0:
+            self._dropout = SpatialDropout1D(self.dropout_rate)
+        else:
+            self._dropout = None
+
+        self.built = True
+
+    def call(self, inputs):
+        # 1. do embedding
+        embedding_output = super(VarLenColumnEmbedding, self).call(inputs)
+
+        # 2. add dropout
+        if self._dropout is not None:
+            dropout_output = self._dropout(embedding_output)
+        else:
+            dropout_output = embedding_output
+
+        # 3. expand dim for polling
+        inputs_4d = tf.expand_dims(dropout_output, 3)  # add channels dim
+
+        # 4. polling
+        tensor_pooling = self._pooling_layer(inputs_4d)
+
+        # 5. format output
+        return tf.squeeze(tensor_pooling, 3)
+
+    def compute_mask(self, inputs, mask):
+        return None
+
+    def get_config(self, ):
+        config = {'pooling_strategy': self.pooling_strategy}
+        base_config = super(VarLenColumnEmbedding, self).get_config()
+        return dict(list(base_config.items()) + list(config.items()))
+
+
 class BinaryFocalLoss(losses.Loss):
     """
     Binary form of focal loss.
@@ -986,7 +1038,7 @@ class CategoricalFocalLoss(losses.Loss):
             gamma {float} -- (default: {2.0})
             alpha {float} -- (default: {4.0})
         """
-        super(BinaryFocalLoss, self).__init__(reduction=reduction, name=name)
+        super(CategoricalFocalLoss, self).__init__(reduction=reduction, name=name)
         self.gamma = float(gamma)
         self.alpha = float(alpha)
 
